%{
open Ast;;


let parse_error s = (* Called by the parser function on error *)
  print_endline s;
  flush stdout
;;

let mknil = Nvar("[]", [])
let mkcons l tl = (List.fold_right (fun t ts -> Nvar("|", [t;ts])) l tl)
%}

%token <string> ID
%token LBRACK RBRACK LPAR RPAR EQ COMMA COLONDASH PERIOD 
%token CONS
%token EOF

%start program
%type <Ast.program> program

%%

program:   clauses goal EOF     { ($1, $2) }

goal:      COLONDASH atom PERIOD { $2 }

clauses:   clause               { [$1] }
         | clause clauses       { $1::$2 }


clause:    atom PERIOD          { [$1] }
         | atom COLONDASH atoms PERIOD
                                { $1::$3 }

atoms:     atom                 { [$1] }
         | atom COMMA atoms     { $1::$3 }

atom:    | ID                   { Nvar($1, []) }
         | ID LPAR atoms RPAR   { Nvar($1, $3) }
         | atom EQ atom         { Nvar("=", [$1;$3]) }
         | LBRACK atoms CONS atom RBRACK { (mkcons $2 $4) }
         | LBRACK atoms RBRACK  { (mkcons $2 mknil) }
         | LBRACK RBRACK        { mknil }
         | LPAR atom RPAR       { $2 }

%%
