open Ast;;
open Util;;
open List;;

let trace_unify = ref false
let trace_call = ref false

exception Unif_fail

let rec unify_list subst t1s t2s = fold_left2 unify subst t1s t2s
and unify subst t1 t2 = match (t1, t2) with
  | (Var(x), _) -> add subst x t2
  | (_, Var(y)) -> add subst y t1
  | (Nvar(c,t1s), Nvar(d,t2s)) -> 
       if c=d then unify_list subst t1s t2s
       else raise Unif_fail

and add subst x t = 
  try let t' = assoc x subst in unify subst t' t
  with Not_found -> if t <> Var(x) then (x,t)::subst else subst
;;

let rec alloc_locals bp ts = 
  let alloc_local t = match t with
    | Var(i) -> Var(bp+i)
    | Nvar(c, ts) -> Nvar(c, alloc_locals bp ts)
  in map alloc_local ts

let assign_to_formals hd goal subst = 
  if (!trace_unify) then (print_string "unify "; print_term hd; print_string " "; (print_term goal); print_string " "; print_env (0, subst); print_string "\n"); 
  unify subst hd goal

let find_body goal (bp, subst) clause: path = 
    let (hd::body) = alloc_locals bp clause in
    try let subst' = assign_to_formals hd goal subst 
        in (body, (bp+(numvars hd)+(numvarslist body), subst'))
    with Unif_fail -> ([], (-1, subst))

let rec call prog env goal = 
  let paths = (map (find_body goal env) prog) in
  let viable_paths = filter (fun (_, (bp, _)) -> bp > 0) paths
  in exec_paths prog viable_paths

and exec_paths prog paths = match paths with
  | [] -> []
  | path1::paths -> (append (exec_path prog path1) (exec_paths prog paths))

and exec_path prog (glist, env) = match glist with
  | [] -> [env]
  | goal::goals -> 
    let envs = call prog env goal in
    let newpaths = map (fun e -> (goals, e)) envs
    in (flatten (map (exec_path prog) newpaths))
;;

let init_env goal = ((numvars goal)+1, [])

let main =
  (* first arg is cflat input file *)
  if ((Array.length Sys.argv) > 2) then 
      (if (String.contains Sys.argv.(2) 'c') then trace_call := true;
       if (String.contains Sys.argv.(2) 'u') then trace_unify := true);
  let (rules, goal) = parse_file Sys.argv.(1) in
  let eqrule: term list = [Nvar("=", [(Var 1);(Var 1)])] in
  let rules' = eqrule::(List.map recognize_varlist1 rules) in
  let (goal', _) = recognize_vars goal [] in
     print_term goal'; print_string "\n"; 
     (print_with_subst (vars goal') (call rules' (init_env goal') goal')); 
     print_string "\n"
